import Globals
from . import virtualgpio as GPIO
import six
import asyncio

from rpcmethods import *


@six.add_metaclass(DevMeta)
class Curtain:
    __devname__  = 'curtains'

    def __init__(self, name, pin_up, pin_down, movement_time, touch_down, touch_up, long_touch_time):
        self.name = name
        self.circuit = parse_name(name)
        self.pin_up = pin_up
        self.pin_down = pin_down
        self.run_time = movement_time
        self.touch_up = touch_up
        self.touch_down = touch_down
        self.long_touch_time = long_touch_time
        self.motor_up = False
        self.motor_down = False
        self.task = None
        self.start_time = None

    def check(self):
        device = check_dev_type(self.pin_up, GPIO.gpout, 'Curtain %s'%self.name, 'pin_up')
        self.pin_up = device
        if device: device.startup_value = 0

        device = check_dev_type(self.pin_down, GPIO.gpout, 'Curtain %s'%self.name, 'pin_down')
        self.pin_down = device
        if device: device.startup_value = 0

        if not(self.touch_up is None):
            device = check_dev_type(self.touch_up, GPIO.gpin, 'Curtain %s'%self.name, 'touch_up')
            self.touch_up = device
            if device: 
                device.add_event_detect(GPIO.RISING, callback=self.touch_up_rise)
                device.add_event_detect(GPIO.FALLING, callback=self.touch_fall)
        if not(self.touch_down is None):
            device = check_dev_type(self.touch_down, GPIO.gpin, 'Curtain %s'%self.name, 'touch_down')
            self.touch_down = device
            if device: 
                device.add_event_detect(GPIO.RISING, callback=self.touch_down_rise)
                device.add_event_detect(GPIO.FALLING, callback=self.touch_fall)

        Globals.logger.debug("Curtain %s initialized",self.name)


    async def touch_up_rise(self, inp):
        loop = asyncio.get_running_loop()
        self.start_time = loop.time()
        await self.open()

    async def touch_down_rise(self, inp):
        loop = asyncio.get_running_loop()
        self.start_time = loop.time()
        await self.close()

    async def touch_fall(self, inp):
        loop = asyncio.get_running_loop()
        if not self.start_time or (self.start_time + self.long_touch_time > loop.time()):
            self.start_time=None
            await self.stop()
        self.start_time=None

    async def open(self):
        if self.task : 
            self.task.cancel()
            self.task = None
        self.task = asyncio.create_task(self._move_up(self.run_time))
        return "OK"

    async def close(self):
        if self.task : 
            self.task.cancel()
            self.task = None
        self.task = asyncio.create_task(self._move_down(self.run_time))
        return "OK"

    async def stop(self):
        if self.task : 
            self.task.cancel()
            self.task = None
        await self.pin_down.output( GPIO.R_OFF)
        await self.pin_up.output( GPIO.R_OFF)
        self.motor_up = False
        self.motor_down = False
        return "OK"

    async def _move_up(self, maxtime):
        if self.motor_down: 
            #make 100% sure we do not destroy the curtain by trying to move it up and down in the same time
            await self.pin_down.output(GPIO.R_OFF)
            await asyncio.sleep(0.25)
            self.motor_down = False
        self.motor_up = True
        await self.pin_up.output(GPIO.R_ON)
        await asyncio.sleep(maxtime)
        await self.pin_up.output(GPIO.R_OFF)
        self.motor_up = False
        self.task = None

    async def _move_down(self, maxtime):
        if self.motor_up:
            #make 100% sure we do not destroy the curtain by trying to move it up and down in the same time
            await self.pin_up.output(GPIO.R_OFF)
            await asyncio.sleep(0.25)
            self.motor_up = False
        self.motor_down = True
        await self.pin_down.output(GPIO.R_ON)
        await asyncio.sleep(maxtime)
        await self.pin_down.output(GPIO.R_OFF)
        self.motor_down = False
        self.task = None



Foptions = {
    'pin_up':       {'type': 'node', 'help': 'Link to some relay' },
    'pin_down':     {'type': 'node', 'help': 'Link to some relay' },
    'movement_time':{'type': float,  'help': 'Max time to move curtain' },
    'touch_up':     {'type': 'node', 'help': 'Link to some input. Can be undefined', 'default': None  },
    'touch_down':   {'type': 'node', 'help': 'Link to some input. Can be undefined', 'default': None  },
    'long_touch_time':   {'type': float, 'help': 'Time to hold down touch sensor to be long', 'default': 2.0  },
}

def YFactory(Node, devname, parent=None):
    kwargs = get_kwargs(Foptions, Node)
    return  Curtain(devname, **kwargs)

'''
  register all remote procedures
'''
def Methods():
    curtains = Curtain.__devlist__
    #      "move curtain completely up/down"
    setattr(RPCMethods,'curtainOpen',rpc_circuit(curtains,'open'))
    setattr(RPCMethods,'curtainClose',rpc_circuit(curtains,'close'))
    #      "start moving curtain up/down -- only on mouse-down/up"
    setattr(RPCMethods,'curtainStartUp',rpc_circuit(curtains,'open'))
    setattr(RPCMethods,'curtainStartDown',rpc_circuit(curtains,'close'))
    #      "called on mouseup"
    setattr(RPCMethods,'curtainStop',rpc_circuit(curtains,'stop'))


# register rpc on using this module
Methods()
