import Globals
from . import virtualgpio as GPIO
import sys
import six
import asyncio

from rpcmethods import *
import broker
#import actions

#GATE STATES
GATE_OPENED = 0
GATE_CLOSED = 1
GATE_OPENING = 2
GATE_CLOSING = 3

FAKE_OPENED = 10
FAKE_CLOSED = 11
FAKE_OPENING = 12
FAKE_CLOSING = 13

CAR_DETECTED = 4

GATE_OPEN_TIME = 10
GATE_RESTRICT_TIME = 3

@six.add_metaclass(DevMeta)
class Gate:

    __devname__  = 'gates'

    def __init__(self, name, close_sensor, open_sensor, control, car_detector, movement_time):
        self.name = name
        self.circuit = parse_name(name)
        self.close_sensor = close_sensor
        self.open_sensor = open_sensor
        self.control = control
        self.car_detector = car_detector
        self.movement_time = movement_time
        self.lock_status = None
        self.restricted = False
        self.task = None
        self.state = None  # unknown

    def check(self):
        if self.control:
            device = check_dev_type(self.control, GPIO.gpout, 'Gate %s'%self.name, 'control')
            self.control = device
            if device: device.startup_value = 0

        if (self.close_sensor):
            device = check_dev_type(self.close_sensor, GPIO.gpin, 'Gate %s'%self.name, 'close_sensor')
            self.close_sensor = device
            if device:
                #if self.is_NO: device.set_reversed()
                device.add_event_detect(GPIO.BOTH, callback=self.close_sensor_event)
                device.add_event_start(callback=self.close_sensor_startup)

        if (self.open_sensor):
            device = check_dev_type(self.open_sensor, GPIO.gpin, 'Gate %s'%self.name, 'open_sensor')
            self.open_sensor = device
            if device:
                #if self.is_NO: device.set_reversed()
                device.add_event_detect(GPIO.BOTH, callback=self.open_sensor_event)
                device.add_event_start(callback=self.open_sensor_startup)

        if (self.car_detector):
            device = check_dev_type(self.car_detector, GPIO.gpin, 'Gate %s'%self.name, 'car_detector')
            self.car_detector = device
            if device:
                #if self.is_NO: device.set_reversed()
                device.add_event_detect(GPIO.FALLING, callback=self.carDetected)

        Globals.logger.debug("Gate %s initialized",self.name)


    def start(self):
        # request setting lock_status from upper system
        broker.callobj('gateRequest',self)

    def sendUpdate(self):
        state = self.state - 10 if self.state >= 10 else self.state
        broker.callobj('gateEvent',self, state, self.lock_status)
        return "OK"

    def getState(self):
        return self.state

    #called when car is detected
    async def carDetected(self, pin):
        Globals.logger.debug("Car detected on Gate %s", self.name)
        broker.callobj('gateEvent',self, CAR_DETECTED, self.lock_status)
        #try:
        self.do_open()
        #except Exception as E:
        #    print("%s %s" %(sys.exc_info()[0],sys.exc_info()[1]))
        #    print(str(E))


    async def close_sensor_startup(self, gpin):
        #check if the gate is closed
        print("Gate %s close sensor=%s" % (self.name, gpin.value))
        if (gpin.value==1):
            self.state = GATE_CLOSED
            Globals.logger.debug("Gate %s closed on init", self.name)
        elif self.state != GATE_OPENED:
            self.state = FAKE_OPENED
            Globals.logger.debug("Gate %s probably opened on init", self.name)

    async def open_sensor_startup(self, gpin):
        #check if the gate is opened
        print("Gate %s open sensor=%s" % (self.name, gpin.value))
        if (gpin.value==1):
            self.state = GATE_OPENED
            Globals.logger.debug("Gate %s opened on init", self.name)
        elif not (self.state in (GATE_CLOSED,FAKE_OPENED)):
            self.state = FAKE_CLOSED
            Globals.logger.debug("Gate %s probably closed on init", self.name)


    async def close_sensor_event(self, gpin):
        Globals.logger.debug("Gate %s %s...", self.name, "closed" if gpin.value == 1 else "opening")
        print("Gate %s %s..." % (self.name, "closed" if gpin.value == 1 else "opening"))
        if gpin.value == 1:
            # confirmed close
            self.state = GATE_CLOSED
            self.clear_timeout()
            self.check_lock()
        elif self.state in (GATE_CLOSED,):
            self.state = GATE_OPENING
            self.plan_timeout(self.movement_time)
        elif self.state in (FAKE_OPENING,):
            self.state = GATE_OPENING
        self.sendUpdate()

    async def open_sensor_event(self, gpin):
        Globals.logger.debug("Gate %s %s...", self.name, "opened" if gpin.value == 1 else "closing")
        print("Gate %s %s..." % (self.name, "opened" if gpin.value == 1 else "closing"))
        if gpin.value == 1:
            # confirmed open
            self.state = GATE_OPENED
            self.clear_timeout()
            self.check_lock()
        elif self.state in (GATE_OPENED,):
            self.state = GATE_CLOSING
            self.plan_timeout(self.movement_time)
        elif self.state in (FAKE_CLOSING,):
            self.state = GATE_CLOSING
        self.sendUpdate()


    def check_lock(self):
        if (self.lock_status==True) and self.state in (GATE_OPENED, FAKE_OPENED):
            self.state = FAKE_CLOSING
            self.plan_timeout(self.movement_time, do_tick=True)
        elif (self.lock_status==False) and self.state in (GATE_CLOSED, FAKE_CLOSED):
            self.state = FAKE_OPENING
            self.plan_timeout(self.movement_time, do_tick=True)


    def do_unlock(self):
        self.lock_status = False
        if self.state == None: 
            self.plan_timeout(self.movement_time)
            return
        self.do_open()
            

    def do_lock(self):
        self.lock_status = True
        if self.state == None: 
            self.plan_timeout(self.movement_time)
            return

        if self.state in (GATE_OPENED, FAKE_OPENED, GATE_OPENING):
            self.state = FAKE_CLOSING
            self.plan_timeout(self.movement_time, do_tick=True)

        elif self.state in (FAKE_CLOSING,):
            if not self.restricted:
                self.plan_timeout(self.movement_time, do_tick=True)

        elif self.state in (FAKE_OPENING,):
            if not self.restricted:
                self.state = FAKE_CLOSING
                self.plan_timeout(self.movement_time, do_tick=True)

    def do_open(self):
        if self.state in (GATE_CLOSED, FAKE_CLOSED):
            self.state = FAKE_OPENING
            self.plan_timeout(self.movement_time+GATE_OPEN_TIME, do_tick=True)

        elif self.state in (FAKE_OPENING,):
            if not self.restricted:
                self.plan_timeout(self.movement_time+GATE_OPEN_TIME, do_tick=True)

        elif self.state in (FAKE_CLOSING, GATE_CLOSING):
            if not self.restricted:
                self.state = FAKE_OPENING
                self.plan_timeout(self.movement_time+GATE_OPEN_TIME, do_tick=True)


    def clear_timeout(self):
        if self.task:
            self.task.cancel()
        self.restricted = False
        self.task = None

    def plan_timeout(self, timeout, do_tick=False):
        if self.task:
            self.task.cancel()
            self.restricted = False
            self.task = None
        self.task = asyncio.create_task(self.tick_task(timeout, do_tick))

    async def tick_task(self, timeout, do_tick = False):
        if do_tick:
            self.sendUpdate()
            Globals.logger.debug("Gate %s is trying to %s", self.name, "close" if self.state in (FAKE_CLOSING,) else "open")
            self.restricted = True
            await self.control.set_for_time(GPIO.R_ON, 2) #0.05)
            await asyncio.sleep(GATE_RESTRICT_TIME)
            timeout -= GATE_RESTRICT_TIME
            self.restricted = False

        if timeout > 0.0:
             await asyncio.sleep(timeout)

        if self.state == GATE_OPENING:
            self.state = FAKE_OPENED

        elif self.state == FAKE_OPENING:
            if self.close_sensor and (self.close_sensor.value == 1):
                self.state = GATE_CLOSED
            else:
                self.state = FAKE_OPENED

        elif self.state == GATE_CLOSING:
            self.state = FAKE_CLOSED

        elif self.state == FAKE_CLOSING:
            if self.open_sensor and (self.open_sensor.value == 1):
                self.state = GATE_OPENED
            else:
                self.state = FAKE_CLOSED

        self.sendUpdate()
        Globals.logger.debug("Gate %s %s(probably)", self.name, "closed" if self.state in (FAKE_CLOSED,GATE_CLOSED) else "opened")
        self.task = None
        self.check_lock()



Foptions = {
    'control':      {'type': 'node', 'help': 'Link to some relay' },
    'close_sensor': {'type': 'node', 'help': 'Link to some input', 'default' : None  },
    'open_sensor':  {'type': 'node', 'help': 'Link to some input', 'default' : None },
    'car_detector': {'type': 'node', 'help': 'Link to some input', 'default' : None },
    'movement_time':{'type': float,  'help': 'Time to move gate' },
}


def YFactory(Node, devname, parent=None):
    kwargs = get_kwargs(Foptions, Node)
    return Gate(devname, **kwargs)

'''
  register all remote procedures
'''
def Methods():
    gates = Gate.__devlist__
    setattr(RPCMethods,'gateGetState',rpc_circuit(gates,'getState'))
    setattr(RPCMethods,'gateSendUpdate',rpc_circuit(gates,'sendUpdate'))
    setattr(RPCMethods,'gateOpen',rpc_circuit(gates,'do_open'))
    setattr(RPCMethods,'gateClose',rpc_circuit(gates,'do_lock'))
    setattr(RPCMethods,'gateLock',rpc_circuit(gates,'do_lock'))
    setattr(RPCMethods,'gateUnlock',rpc_circuit(gates,'do_unlock'))


# register rpc on using this module
Methods()

