import Globals
from . import virtualgpio as GPIO
import six

from rpcmethods import *
import broker
import server

@six.add_metaclass(DevMeta)
class Heating:

    __devname__  = 'heatings'
    faketemp = 20

    def __init__(self, name, pin, desired, antifreeze_setpoint, is_season, sensor):
        self.name = name
        self.circuit = parse_name(name)
        self.pin = pin
        self.desired_temp = desired
        self.antifreeze_setpoint = antifreeze_setpoint
        self.is_season = is_season
        self.sensor = sensor
        self.window_opened = False
        self.sleep_timer = 0
        #self.actual_temp = 0
        #self.target_temp = 0
        self.setpoint = desired
        self.is_heating = None
        self.task = None

    def check(self):
        device = check_dev_type(self.pin, GPIO.gpout, 'Heating %s'%self.name, 'pin')
        self.pin = device
        device.add_callback(self.relayEvent)
        #device.startup_value = 0
        self.pin = device
        if self.sensor:
            device = check_dev_type(self.sensor, GPIO.Thermometer, 'Heating %s'%self.name, 'sensor')
            self.sensor = device
        Globals.logger.debug("Heating %s initialized", self.name)

    def sendUpdate(self):
        broker.callobj('heatingEvent',self,self.getTemp(),self.setpoint,self.antifreeze_setpoint, self.is_heating, self.is_season)
        return "OK"

    async def relayEvent(self, value):
        ''' acknoledgment of relay status - called from gpout'''
        self.is_heating = value
        #broker.callobj('lightEvent',self, self.light_status)

    def getTemp(self, fake = True):
        self.actual_temp = 20
        if self.sensor:
            actual_temp = self.sensor.get_valid(300)
            if not(actual_temp is None):
                return actual_temp
        if fake: return self.faketemp
        return None

    def setDesired(self, temp):
        self.desired_temp = temp
        self.setpoint = self.antifreeze_setpoint if self.window_opened else self.desired_temp
        return "OK"

    def getDesired(self):
        return self.desired_temp

    def setWindowOpenedTemp(self, temp):
        self.window_opened_temp = temp
        self.setpoint = self.antifreeze_setpoint if self.window_opened else self.desired_temp
        return "OK"

    def getWindowOpenedTemp(self):
        return self.antifreeze_setpoint

    def setSeason(self, is_season):
        self.is_season = is_season in (True, 1, '1', 'True')
        if self.is_season and not self.task:
            self.task = asyncio.create_task(self.run())

        #elif not self.is_season and self.task:  # not needed - finish on timer in loop
        #    self.task.cancel()
        #    self.task = None
        return "OK"

    def getSeason(self):
        return self.is_season

    def setWindowOpened(self, is_open):
        self.window_opened = is_open in (True, 1, '1', 'True')
        self.setpoint = self.antifreeze_setpoint if self.window_opened else self.desired_temp
        return "OK"

    def getWindowOpened(self):
        return self.window_opened


    def start(self):
        async def create_task(app):
            self.task = asyncio.create_task(self.run())
        server.register_background_task(create_task)


    async def run(self):

        for _ in range(20):
            if not(self.is_heating is None) and self.getTemp(fake=False): break
            await asyncio.sleep(1)

        while self.is_season:
            actual_temp = self.getTemp()
            if (actual_temp < self.setpoint - 0.5) and not(self.is_heating):
                await self.pin.output( GPIO.R_ON)
                Globals.logger.debug("Heating - low temp(%.1f), heating to %.1f", actual_temp, self.setpoint)
                await asyncio.sleep(600)
            elif ((actual_temp <= self.setpoint + 0.5) and (prev_heat == False)):
                await self.pin.output( GPIO.R_ON)
                Globals.logger.debug("Heating - good temp(%.1f), keeping - %.1f", actual_temp, self.setpoint)
                await asyncio.sleep(300)
            else:
                await self.pin.output( GPIO.R_OFF)
                Globals.logger.debug("Heating - high temp(%.1f), cooling to %.1f", actual_temp, self.setpoint)
                await asyncio.sleep(900)

        self.task = None
        await self.pin.output( GPIO.R_OFF)
        self.sendUpdate()


Foptions = {
    'pin':       {'type': 'node', 'help': 'Link to some relay - valve' },
    'sensor':    {'type': 'node', 'help': 'Link to some temperature sensor' },
    'desired':   {'type': float,  'help': 'Setpoint temperature on normal condition [23.0]', 'default': 23 },
    'antifreeze_setpoint':{'type': float,  'help': 'Temperature to be during window opened. [15.0]', 'default': 15 },
    'is_season':{'type': int,  'help': 'Is heating season. Default 1', 'default': 1 },
}


def YFactory(Node, devname, parent=None):
    kwargs = get_kwargs(Foptions, Node)
    return Heating(devname, **kwargs)


'''
  register all remote procedures
'''
def Methods():
    heatings = Heating.__devlist__
    setattr(RPCMethods,'heatingSendUpdate',rpc_circuit(heatings,'sendUpdate'))
    setattr(RPCMethods,'heatingGetDesiredTemp',rpc_circuit(heatings,'getDesired'))
    setattr(RPCMethods,'heatingGetWindowOpenedTemp',rpc_circuit(heatings,'getWindowOpenedTemp'))
    setattr(RPCMethods,'heatingGetSeason',rpc_circuit(heatings,'getSeason'))
    setattr(RPCMethods,'heatingGetWindowOpened',rpc_circuit(heatings,'getWindowOpened'))
    setattr(RPCMethods,'heatingSetDesiredTemp',rpc_circuit_one(heatings,'setDesired'))
    setattr(RPCMethods,'heatingSetWindowOpenedTemp',rpc_circuit_one(heatings,'setWindowOpenedTemp'))
    setattr(RPCMethods,'heatingSetSeason',rpc_circuit_one(heatings,'setSeason'))

# register rpc on using this module
Methods()
