import Globals
from . import virtualgpio as GPIO
import six
import asyncio

from rpcmethods import *
import broker

@six.add_metaclass(DevMeta)
class SlidingDoor:
  ## this class shares __devname__ and methods with Doors

  __devname__  = 'doors'

  def __init__(self, name, open_pin, lock_pin, close_sensor):
        self.name = name
        self.circuit = parse_name(name)
        self.open_pin = open_pin
        self.lock_pin = lock_pin
        self.close_sensor = close_sensor
        self.is_locked = None
        self.task = None

  def check(self):
        device = check_dev_type(self.open_pin, GPIO.gpout, 'Sliding Door %s'%self.name, 'open_pin')
        self.open_pin = device
        if device: device.startup_value = GPIO.R_OFF

        device = check_dev_type(self.lock_pin, GPIO.gpout, 'Sliding Door %s'%self.name, 'lock_pin')
        if device: device.startup_value = GPIO.R_OFF
        self.lock_pin = device

        device = check_dev_type(self.close_sensor, GPIO.gpin, 'Sliding Door %s'%self.name, 'close_sensor')
        self.close_sensor = device
        if device:
            #if self.is_NO: device.set_reversed()
            device.add_event_detect(GPIO.BOTH, callback=self.doorEvent)
            device.add_event_start(callback=self.doorStartupEvent)

        Globals.logger.debug("Sliding Door %s initialized",self.name)

  async def doorStartupEvent(self, gpin):
        self.is_opened = self.close_sensor.value
        if (self.is_opened):
            self.is_locked = False

  def sendUpdate(self):
        broker.callobj('doorEvent',self,self.is_opened, self.is_locked)
        return "OK"

  def getState(self):
        return (self.is_opened, self.is_locked)

  async def doorEvent(self, gpin):
        self.is_opened = self.close_sensor.value
        if (self.is_opened):
            if self.is_locked is None : self.is_locked = False

        Globals.logger.debug("Sliding Door %s %s", self.name, "opened" if self.is_opened else "closed")
        self.sendUpdate()

  async def open(self):
        await self.open_pin.set_for_time(GPIO.R_ON, 0.1)
        return "OK"


  async def _trylock(self, do_lock):
        GLobals.logger.debug("Trying to check locking state door %s", self.name)
        self.trying_lock = True
        if (self.is_locked is None):
            await self.lock_pin.set_for_time(GPIO.R_ON, 0.1)
            await asyncio.sleep(5) #wait for door opening. if opened intime, is_locked is set
            if self.is_locked is None:
                self.is_locked = True
                Globals.logger.debug("Sliding door %s locked", self.name)
        else:
            # colision, somebody solved the problem
            pass

        if do_lock and not self.is_locked:
            await self.lock_pin.set_for_time(GPIO.R_ON, 0.1)
            self.is_locked = True
            Globals.logger.debug("Sliding door %s locked", self.name)

        elif not do_lock and self.is_locked:
            await self.lock_pin.set_for_time(GPIO.R_ON, 0.1)
            self.is_locked = True
            Globals.logger.debug("Sliding door %s locked", self.name)
        self.task = None


  async def lock(self):
        if  self.task:
            return "Just trying to check locking state"

        if not (self.is_locked):
            self.task = asyncio.create_task(self._trylock(True))
        return "OK"

  async def unlock(self):
        if  self.task:
            return "Just trying to check locking state"

        if (self.is_locked is None) or self.locked:
            self.task = asyncio.create_task(self._trylock(False))
        return "OK"


Foptions = {
    'open_pin':     {'type': 'node', 'help': 'Link to some relay' },
    'lock_pin':     {'type': 'node', 'help': 'Link to some relay' },
    'close_sensor': {'type': 'node', 'help': 'Link to some input' },
}


def YFactory(Node, devname, parent=None):
    kwargs = get_kwargs(Foptions, Node)
    return SlidingDoor(devname, **kwargs)


'''
  register all remote procedures
'''
def Methods():
    # methods are share with Doors, should be same names
    doors = SlidingDoor.__devlist__
    setattr(RPCMethods,'doorGetState',rpc_circuit(doors,'getState'))
    setattr(RPCMethods,'doorSendUpdate',rpc_circuit(doors,'sendUpdate'))
    setattr(RPCMethods,'doorOpen',rpc_circuit(doors,'open'))
    setattr(RPCMethods,'doorLock',rpc_circuit(doors,'lock'))
    setattr(RPCMethods,'doorUnlock',rpc_circuit(doors,'unlock'))


# register rpc on using this module
Methods()
