#!/usr/bin/python3
import os
import sys
import re
import time
import asyncio
from queue import Queue
import aiohttp

from future.utils import iteritems

import logging
import logging.handlers

from rpcmethods import *
import broker
import server

############################## INI Configuration ###################################

def LoadIniConfig(path):

    import ConfigParser
    Config = ConfigParser.SafeConfigParser()
    Globals.logger.info('Reading INI configuration')
    Config.read(path)
    # transform Ini to Yaml like dictionary
    return { section.lower(): { par: val for par,val in Config.items(section)} for section in Config.sections() }


############################## YAML Configuration ###################################
def LoadYamlConfig(path):
    import yaml
    with open(path, 'r') as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as E:
            Globals.logger.critical("Yaml error: %s", str(E))
            raise SystemExit() 


def LoadNode(section, node, parent):
    if type(node) is list:
        for item in node:
            LoadNode(None, item, parent)
        return

    if not(type(node) is dict):
        Globals.logger.critical("Bad config %s", str(node))
        return

    if section is None:
        for isection, item in iteritems(node):
            LoadNode(isection, item, parent)
        return

    # parse type attribute and Node name
    if section.startswith('_'): return  # skip disabled items
    prefix = 'device'
    devicetype = node.get("type", '')
    if not devicetype:
        res = re.search('^(.*\D)(\d+)$',section)
        if res: devicetype=res.group(1)
    if '.' in devicetype:
        prefix,devicetype = devicetype.split('.',1)

    #print(prefix, devicetype, section)
    try:
        module = __import__("%s.%s" % (prefix,devicetype), fromlist=[prefix,])
    except ImportError:
        Globals.logger.critical("Config node '%s': Unknown device type=%s", section, devicetype)
        if Globals.deny_config_errors: raise
        return

    obj = None
    if hasattr(module,'YFactory'):
        try:
            obj = module.YFactory(node, section, parent)
        except (AttributeError, KeyError) as E:
            Globals.logger.critical("Config node '%s': Missing required parameter %s",section, str(E))
            if Globals.deny_config_errors: raise
        except Exception:
            Globals.logger.critical("Config node' %s': %s %s", section, sys.exc_info()[0],sys.exc_info()[1])
            if Globals.deny_config_errors: raise

    childs = node.get('childs', None)
    if childs: LoadNode(None, childs, obj)


############################## Queue Logging ###################################
def setup_logging_queue() -> None:
    """Move log handlers to a separate thread.

    Replace handlers on the root logger with a LocalQueueHandler,
    and start a logging.QueueListener holding the original
    handlers.

    """
# QueueHandler, QueueListener from logging.handlers
    queue = Queue()
    #root = logging.getLogger()

    handlers: List[logging.Handler] = []

    handler = logging.handlers.QueueHandler(queue)
    logger = Globals.logger.addHandler(handler)
    for h in Globals.logger.handlers[:]:
        if h is not handler:
            Globals.logger.removeHandler(h)
            handlers.append(h)

    listener = logging.handlers.QueueListener(
        queue, *handlers, respect_handler_level=True
    )
    listener.start()


def main(args):
    #setup logging
    logger = Globals.logger 
    logger.setLevel(Globals.SYSLOG_LEVEL)
    if True:
        # add console logger in simple message format
        fh = logging.StreamHandler(sys.stdout)
        fh.setLevel(logging.DEBUG if args.debug else logging.INFO)
        if args.debug: logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)-8s: %(message)s')
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        #logger2 = logging.getLogger('pymodbus.client.asynchronous.async_io')
        #logger2.addHandler(fh)
        #if args.debug: logger2.setLevel(logging.DEBUG)

    if args.syslog:
        # add syslog logger in decorated format
        #fh = logging.FileHandler('/var/log/'+Globals.APP_NAME+'.log')
        fh = logging.handlers.SysLogHandler(address='/dev/log',	facility='local0')
        fh.setLevel(Globals.SYSLOG_LEVEL)
        formatter = logging.Formatter('%(asctime)s - %(name)s %(levelname)s: %(message)s')
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    Globals.deny_config_errors = args.break_on_error;

    #read config from path specified on command line or from /etc or from program directory
    if args.config:
        path = args.config
    else:
        path = Globals.default_config
        if not os.path.isfile(path):
            path = Globals.run_path +'/'+ os.path.basename(Globals.default_config)

    if not os.path.isfile(path):
        raise Exception('Missing config file %s', path)

    logger.debug("Loading config file: %s", path)

    if path.endswith('.conf'):
        node = LoadIniConfig(path)
    else:
        node = LoadYamlConfig(path)
    LoadNode(None, node, None)

    ###################################################
    # start Queue logging - to be asyncio compatible
    setup_logging_queue()


    ###################################################
    # start brokers first
    for k,unit in iteritems(broker.brokers):
        try_to_start(unit)

    # check all units
    for k,unit in iteritems(Units):
        try:
           if hasattr(unit, 'check'): unit.check()
        except:
            logger.critical("Check unit '%s': %s %s", k, sys.exc_info()[0],sys.exc_info()[1])
            if Globals.deny_config_errors: raise

    # start all units
    for k,unit in iteritems(Units):
        try:
            try_to_start(unit)
        except:
            logger.critical("Start unit '%s': %s %s", k, sys.exc_info()[0],sys.exc_info()[1])
            if Globals.deny_config_errors: raise

    #start RPC server
    try:
        server.serve_forever()
        #while threading.active_count() > 1:
        #    time.sleep(2)
    except (KeyboardInterrupt, asyncio.CancelledError):
        print("SHUTDOWN")
        server.shutdown()
    

if __name__ == '__main__':
    import argparse
    a = argparse.ArgumentParser(prog='homevok.py')
    a.add_argument('-l','--syslog',action='store_true', help='use syslog for logging')
    a.add_argument('-d','--debug',action='store_true', help='debug logging on stdout')
    a.add_argument('-b','--break',action='store_true', help='break program on exception (during config)', dest='break_on_error')
    a.add_argument('-c','--config',help='use this config file (default: homevok.conf from /etc or program directory)',
                                   metavar="file")
    args = a.parse_args()
    try:
        main(args)

    except Exception as E:
        Globals.logger.critical("Error %s %s", sys.exc_info()[0],sys.exc_info()[1])
        import traceback
        traceback.print_exc(file=sys.stdout)
 
    os._exit(1)
    os.kill(0,1)
