import Globals
import logging
import threading
import time
import datetime
import sys
import six

from pymodbus.client.asynchronous.tcp import AsyncModbusTCPClient as ModbusTcpClient
from pymodbus.client.asynchronous import schedulers
from pymodbus.client.asynchronous.async_io import  ReconnectingAsyncioModbusTcpClient

from rpcmethods import *
from device import virtualgpio as GPIO
import server
from . import get_client_loop

class BaseModbusClient(object):
    pass

@six.add_metaclass(DevMeta)
class ModbusClient(BaseModbusClient):

    def __init__(self, name, host, port):
        self.name = name
        self.circuit = parse_name(name)
        self.host = host
        self.port = port
        self.connection = None
        self.device_event=dict()
        self.running_lock = asyncio.Lock()

    def start(self):
        async def create_task(app):
            asyncio.create_task(self.run())
        async def stop_task(app):
            asyncio.create_task(self.stop())
        server.register_background_task(create_task, stop_task)

    async def stop(self):
        if self.connection: self.connection.stop()

    async def run(self):
        Globals.logger.info("Starting Modbus client %s:%s", self.host, self.port)
        proto_cls = None #kwargs.get("proto_cls", None)
        client = ReconnectingAsyncioModbusTcpClient(protocol_class=proto_cls, loop=asyncio.get_running_loop())
        await client.start(self.host, self.port)
        Globals.logger.debug("Modbus client %s:%s %s", self.host, self.port, "connected" if client.connected else "not connected")
        self.connection = client
        if not self.connection.connected : return
        #reg = await self.connection.protocol.read_input_registers(1000,count=4,unit=1)
        #print(reg)
        #print(reg.registers)

    async def _run(self):
        print("Starting Modbus client", self.host, self.port)
        loop, connection = ModbusTcpClient(schedulers.ASYNC_IO, host=self.host, port=self.port, loop=get_client_loop())
        self.connection = connection
        if not connection.connected: return
        coro = self.connection.protocol.read_input_registers(1000,count=4,unit=1)
        future = asyncio.run_coroutine_threadsafe(coro, loop)
        timeout=10
        try:
            reg = future.result(timeout)
            print(reg.registers)
        except asyncio.TimeoutError:
            print('The coroutine took too long, cancelling the task...')
            future.cancel()
        except Exception as E:
            print(f'The coroutine raised an exception: {E!r}')


    async def write_coil(self, address, value, unit):
        if self.connection is None : return None
        async with self.running_lock:
            try:
                return await self.connection.protocol.write_coil(address, value, unit=unit)
            except Exception as E:
                print(f'Modbus write coil failed with exception: {E!r}')

    async def read_coils(self, address, count, unit):
        if self.connection is None : return None
        async with self.running_lock:
            try:
                return await self.connection.protocol.read_coils(address, count=count, unit=unit)
            except Exception as E:
                print(f'Modbus read coils {address} failed with exception: {E!r}')
                return None

    async def read_bits(self, address, count, unit):
        if self.connection is None : return None
        async with self.running_lock:
            try:
                return await self.connection.protocol.read_discrete_inputs(address, count=count, unit=unit)
            except Exception as E:
                print(f'Modbus read bits {address} failed with exception: {E!r}')
                return None


Foptions = {
    'host':     {'type': str, 'help': 'Host or ip address' },
    'port':     {'type': int, 'help': 'Tcp port to connect to. Default 502', 'default': 502 },
}


def YFactory(Node, devname, parent = None):
    kwargs = get_kwargs(Foptions, Node)
    return ModbusClient(devname, **kwargs)
