import Globals
import time
import datetime
import sys
import threading
import six

from rpcmethods import *
from .client import BaseModbusClient
from device import virtualgpio as GPIO
import server

@six.add_metaclass(DevMeta)
class ModbusReadGroup():

    #__devname__  = 'inputs'
    errorsleep = 1

    def __init__(self, name, channel, mode, unit, reg, count):
        self.name = name
        self.circuit = parse_name(name)
        self.channel = channel
        self.mode = mode
        self.unit = unit
        self.reg = reg
        self.count = count
        self.interval = 0.1
        self.callbacks = [None] * count


    def check(self):
        device = check_dev_type(self.channel, BaseModbusClient, 'ReadGroup %s'%self.name, 'channel')
        if not self.mode in ("coil", "bit", "holding", "input"):
            Globals.logger.critical("Bad mode (%s) for readgroup %s", self.mode, self.mode)
            device = None;
        self.channel = device


    def register_callback(self, index, callback):
        if (index >= 0) and (index < self.count):
            self.callbacks[index] = callback


    def start(self):
        async def create_task(app):
            if self.mode == "coil" or self.mode == "bit":
                asyncio.create_task(self.run_coil())
            elif self.mode == "holding" or self.mode == "input":
                asyncio.create_task(self.run_coil())

        if self.channel:
            server.register_background_task(create_task)


    async def run_coil(self):
        lastbits = [ None ] * self.count
        while True:
            if self.mode == 'coil':
                res = await self.channel.read_coils(self.reg, self.count, self.unit)
            else:
                res = await self.channel.read_bits(self.reg, self.count, self.unit)
            if (not res) or res.isError():
                if res:
                    Globals.logger.debug("Error read modbus group %s (%s)", self.name, str(res))
                await asyncio.sleep(self.errorsleep)
                continue

            for i in range(self.count):
                if self.callbacks[i] and (lastbits[i] != res.bits[i]):
                    funcres = self.callbacks[i](res.bits[i])
                    if asyncio.iscoroutine(funcres): asyncio.create_task(funcres)
            lastbits = res.bits
            await asyncio.sleep(self.interval)


Foptions = {
    'channel':  {'type': 'node', 'help': 'Link to Modbus client' },
    'mode':     {'type': str, 'help': 'Modbus operation used to read ("coil","bit","holding","input")' },
    'unit':     {'type': int, 'help': 'Modbus device address. Default 0', 'default': 0 },
    'reg':      {'type': int, 'help': 'Modbus register address (starting from 0)' },
    'count':    {'type': int, 'help': 'Modbus register count' },
}

def YFactory(Node, devname, parent = None):
    kwargs = get_kwargs(Foptions, Node, parent)
    return ModbusReadGroup(devname, **kwargs)
