#!/usr/bin/python
import serial_asyncio as serial
import asyncio


def UINT32_T(x:int):
    return (x & 0xffffffff)


def ParseEvent(frame):
    if not frame  or len(frame) < 3: return None
    address = frame[0]
    length = frame[1]
    command = frame[2]
    params =  frame[3:length-2]
    if command != 123:
        return None
    if params[6] & 0x80:
        # Event
        return None, 1
    else:
        code = "%02x%02x%02x%02x%02x" % (params[7],params[8],params[9], params[10],params[11])
        if code [-2:] == '00' : code = code[:-2] 
        # card event
        return code, None


def ParseQueueFrame(frame):
    address = frame[0]
    pointer = frame[5] + (frame[6] << 8)
    capacity = frame[3] + (frame[4] << 8)
    counter = frame[7] | (frame[8] << 8) | (frame[9] << 16) | (frame[10] << 24)
    return (address, pointer, capacity, counter)


def LicxCRC2(bytes):

    result = 0;
    for B in bytes: 
        C=((result >> 8) ^ B) << 8
        for i in range(8):
            if (C & 0x8000): C = (C << 1) ^ 0x1021
            else: C = C << 1
        result = C ^ (result << 8)
    return result & 0xFFFF


COMMAND_TIMEOUT   = 0.05

class NetronixProtocol(asyncio.Protocol):

    BROADCAST_ADDR    = 255
    BROADCAST_TIMEOUT = 0.2

    def __init__(self):
        super().__init__()
        self._transport = None
        self.do_drop_data = True
        self.frameEvent = asyncio.Event()

    def connection_made(self, transport):
        self._transport = transport
        self.do_drop_data = True
        #print('port opened', self._transport,Serial)
        #self._transport.write(b'Hello, World!\n')

    def connection_lost(self, exc):
        #print('port closed')
        self.do_drop_data = True
        #self._transport.loop.stop()

    def data_received(self, data):
        ''' callback from loop 
            - collect data in self._frame
            - emits whole frame using frameEvent
        '''
        #print('data received', repr(data))
        if self.do_drop_data:
            # unexpected data
            return
        self._frame += data
        if len(self._frame) > 1:
            length = self._frame[1]
            if length <= len(self._frame):
                crc = (self._frame[length-2] << 8) + self._frame[length-1]
                if crc == LicxCRC2(self._frame[:length-2]):
                    self.receivedframe = self._frame
                else:
                    self.receivedframe = None # crc error
                self._frame = self._frame[length:]
                self.frameEvent.set()

    async def receive_frame(self, timeout, expect):
        self.frameEvent.clear()
        self._frame = []
        self.do_drop_data = False
        result = None
        try:
            await asyncio.wait_for(self.frameEvent.wait(), timeout=timeout)
            self.frameEvent.clear()
            if self.receivedframe and (self.receivedframe[2] == expect):
                result = self.receivedframe
        except asyncio.TimeoutError:
            pass
        self.do_drop_data = True
        return result

    async def receive_multiframe(self, timeout, expect):
        self._frame = []
        self.frameEvent.clear()
        self.do_drop_data = False
        result = []
        try:
            async def multievent():
                while True:
                    await self.frameEvent.wait()
                    self.frameEvent.clear()
                    if not self.receivedframe or (self.receivedframe[2] != expect):
                        return
                    result.append(self.receivedframe)

            await asyncio.wait_for(multievent(), timeout=timeout)
        except asyncio.TimeoutError:
            pass
        self.do_drop_data = True
        return result


    async def do_command(self, address, command, params, timeout = COMMAND_TIMEOUT):
        length = 5 + len(params)
        frame = [address, length, command] + params
        crc = LicxCRC2(frame)
        frame = frame + [ crc >> 8, crc & 0xff ]
        self._transport.write(bytes(frame))
        #await self._transport.drain()
        if address == self.BROADCAST_ADDR: 
            return await self.receive_multiframe(timeout, command+1)

        return await self.receive_frame(timeout, command+1)


    ##### High level interface ##############

    async def ScanBus(self):
        ''' return dict { 'address' -> EventQueueStatus }
                EventQueueStatus is tuple (address, pointer, capacity, counter)
        '''
        frames = await self.do_command(self.BROADCAST_ADDR, 0x78, [], timeout=self.BROADCAST_TIMEOUT)

        if not frames: return {}
        return { str(frame[0]):ParseQueueFrame(frame) for frame in frames } 

    async def GetEvent(self, address, pointer):
        ''' read one stored event from device(address) from internal position '''
        frame = await self.do_command(address, 0x7a, [pointer & 0xff, (pointer >> 8) & 0xff])
        return ParseEvent(frame)


    async def CheckEventStatus(self, address):
        frame = await self.do_command(address, 0x78, [])
        if not frame: return
        return ParseQueueFrame(frame)


    async def ReadNewEvents(self, EventQueueStatus, obj):
        ''' read first valid result or None from Event queue '''
        #print(EventQueueStatus)
        address, pointer, capacity, counter = EventQueueStatus
        results = []
        if not hasattr(obj, 'nextcounter'):
            obj.nextcounter = UINT32_T(counter + 1)
            return results

        while obj.nextcounter <= counter:
            eventptr = (pointer - (counter-obj.nextcounter+1)) % capacity
            obj.nextcounter = UINT32_T(obj.nextcounter + 1)
            result =  (await self.GetEvent(address, eventptr)) or (await self.GetEvent(address, eventptr))
            if result: 
                results.append(result)

        obj.nextcounter = UINT32_T(counter + 1)
        return results


    async def SetIO(self, address, io, value):
        await self.do_command(address, 0x70, [io, (1 if value else 0)])


'''
async def myfun(reader, writer):
    writer.write(b'Ciao!!')
    await writer.drain()
    #try:
    x = await asyncio.wait_for(reader.read(100), timeout=0.1)
    print(x)
    #except asyncio.TimeoutError:
    #    print("oops took longer than 5s!")
    print('konec')


async def start():
    reader, writer = await serial.open_serial_connection(url='socket://localhost:7000', baudrate=115200)
    #coro = writer.transport
    #asyncio.create_task(coro)
    asyncio.create_task(myfun(reader, writer))
    await asyncio.sleep(1000)


if __name__ == '__main__':

    class Output(asyncio.Protocol):

        def __init__(self):
            super().__init__()
            self._transport = None

        def connection_made(self, transport):
            self._transport = transport
            print('port opened', self._transport)
            #print(self._transport.serial.timeout)
            #self._transport.serial.timeout = None
            #self._transport.serial._write_timeout = None
            #print(self._transport.serial.timeout)
            #self._transport.serial.rts = False
            self._transport.write(b'Hello, World!\n')

        def data_received(self, data):
            print('data received', repr(data))

            #if b'\n' in data:
            #    self._transport.close()

        def connection_lost(self, exc):
            print('port closed')
            self._transport.loop.stop()

        def pause_writing(self):
            print('pause writing')
            print(self._transport.get_write_buffer_size())

        def resume_writing(self):
            print(self._transport.get_write_buffer_size())
            print('resume writing')


    loop = asyncio.get_event_loop()
    #coro = serial.create_serial_connection(loop, Output, 'socket://localhost:7000', baudrate=115200)
    #loop.run_until_complete(coro)
    loop.run_until_complete(start())
    loop.run_forever()
    loop.close()

'''