import Globals
import os
import six
import asyncio
import json

from rpcmethods import *
import server
import broker
import actions

@six.add_metaclass(DevMeta)
class OwBus:

    def __init__(self, name, interval, scan_interval, bus):
        self.name = name
        self.circuit = parse_name(name)
        self.interval = interval
        self.scan_interval = scan_interval
        self.bus = bus
        self.writer = None
        self.callback = {}
        self.persistent = {}


    async def stop(self):
        self.do_stop = True
        await self.send([7,0,0])


    def start(self):
        async def create_task(app):
            asyncio.create_task(self.run())
        async def stop_task(app):
            asyncio.create_task(self.stop())
        server.register_background_task(create_task, stop_task)
    

    async def send(self, data):
        if self.writer:
            self.writer.write(bytes(json.dumps(data)+'\n', encoding='utf-8'))
            await self.writer.drain()

    async def emit(self, data):
        #print(data)
        try:
            (address, typ, value) = data
            if type(value) is bool:
                if value == True: 
                    Globals.logger.debug('OwBus %s lost thermometer %s', self.name, address)
                else:
                    Globals.logger.debug('OwBus %s new thermometer %s', self.name, address)
                return
            value = float(value)
            stime = asyncio.get_running_loop().time()
            if str(address) in self.persistent:
                self.persistent[str(address)] = {'value' : value, 'time':stime}
            if str(address) in self.callback:
                for fun in self.callback[str(address)]:
                    fun(value)
            data = {str(address): (typ, value, stime) }
            broker.call('owEvent', data)
        except Exception as E:
            Globals.logger.debug("OwBus %s error=%s", self.name, str(E))
            return


    async def run(self):
        #PYTHON = os.environ['_']
        cmd = "/usr/bin/python3" 
        args = ("owclient/owprocess.py","-i",str(self.interval),"-s",str(self.scan_interval), self.bus)
        self.do_stop = False
        while not self.do_stop:

            proc = await asyncio.create_subprocess_exec(cmd, *args, stdout=asyncio.subprocess.PIPE, stdin=asyncio.subprocess.PIPE)
            Globals.logger.debug("OwBus %s initialized. pid=%s", self.name, proc.pid)
            self.writer = proc.stdin

            bline = b''
            while not self.do_stop:
                try:
                    bline = await proc.stdout.readuntil()
                    if bline:
                        await self.emit(json.loads(bline.decode()))
                except asyncio.streams.IncompleteReadError:
                    print('closed')
                    break
            try:
                self.writer = proc.stdin
                await asyncio.wait_for(proc.wait(), timeout=2.0)
            except asyncio.TimeoutError:
                proc.terminate()


    def register_callback(self, address, callback):
        try:
            cblist = self.callback[str(address)]
            cblist.append(callback)
        except KeyError:
            self.callback[str(address)] = [callback,]

    def register_persistent(self, address):
        if not (str(address) in self.persistent):
            self.persistent[str(address)] = None

    def get_persistent(self, address):
        try:
            return self.persistent[str(address)]
        except KeyError:
            return None



Foptions = {
    'bus': {'default':'--i2c=/dev/i2c-1:ALL', 'help': 'owserver bus connection string' },
    'interval':      {'default':3, 'type': int, 'help': 'interval[s] between measurement. default 3'},
    'scan_interval': {'default':300, 'type': int, 'help': 'interval[s] between scanning the bus'},
}

def YFactory(Node, devname, parent=None):
    kwargs = get_kwargs(Foptions, Node)
    return OwBus(devname, **kwargs)

