#!/usr/bin/python3
import multiprocessing
import os, sys, select
import time
import onewire as ow
import signal
import string
import logging
import json

MAX_LOSTINTERVAL = 300  # 5 minutes

OWCMD_INTERVAL = 1
OWCMD_SCAN = 2
OWCMD_SCAN_INTERVAL = 3
OWCMD_DEFAULT_INTERVAL = 4
OWCMD_SET_PIO = 5
OWCMD_RESET_MASTER = 6
OWCMD_STOP = 7
SUPPORTED_DEVICES = ["DS18S20", "DS18B20", "DS2438", "DS2408", "DS2413"]

logger = logging.getLogger()

class MySensor(object):
    def __init__(self, address, typ, interval, dynamic=True, is_static=False):
        self.type = typ
        self.address = address
        self.interval = interval  # seconds
        self.is_dynamic_interval = dynamic  # dynamically change interval #TODO
        self.last_value = None
        self.value = None
        self.lost = False
        self.time = 0
        self.readtime = 0
        self.sens = None
        self.is_static = is_static
        self.lost = False

    def read_val_from_sens(self, sens):
        raise NotImplementedError("Please Implement this method")

    def calc_interval(self):
        if self.lost: # Sensor is inactive (disconnected)
            self._lostinterval *= min(self._lostinterval*2, MAX_LOSTINTERVAL)
            return self._lostinterval

        if self.is_dynamic_interval:
            ##TODO
            pass
        return self.interval

    def set_lost(self):
        if not self.lost:
            self.lost = True
            self._lostinterval = self.interval


class DS18B20(MySensor):  # thermometer

    def read_val_from_sens(self, sens):
        try:
            new_val = float(sens.temperature)
        except (AttributeError, TypeError):
            raise AttributeError('Bad value')

        if (new_val == 85.0) and (abs(new_val - self.value) > 2):
            logger.debug("PoR detected! 85C")
            return
        self.value = round(new_val * 2, 1) / 2  # 4 bits for frac part of number


class DS2438(MySensor):  # vdd + vad + thermometer

    def read_val_from_sens(self, sens):
        self.value = (sens.VDD, sens.VAD, sens.temperature, sens.vis)


class DS2408(MySensor):

    def read_val_from_sens(self, sens):
        #actual values must be read from sensed_ALL, but writes to the GPIOs must be done in PIO_x(PIO_ALL)
        #pios_values = map(int, self.sens.sensed_ALL.split(','))
        pios_values = [int(not int(i)) for i in self.sens.sensed_ALL.split(',')]
        self.value = pios_values


def MySensorFabric(address, typ, interval, dynamic=True, is_static=False):
    if (typ == 'DS18B20') or (typ == 'DS18S20'):
        return DS18B20(address, typ, interval, is_static=is_static)
    elif (typ == 'DS2438'):
        return DS2438(address, typ, interval, is_static=is_static)
    elif (typ == 'DS2408') or (typ == 'DS2406') or (typ == 'DS2413'):
        return DS2408(address, typ, interval, is_static=is_static)
    else:
        logger.debug("Unsupported 1wire device %s (%s) detected", typ, address)
        return None


class OwBusDriver(multiprocessing.Process):

    def __init__(self, taskPipe, resultPipe, interval=3, scan_interval=300, bus='--i2c=/dev/i2c-1:ALL'):
        multiprocessing.Process.__init__(self)
        self.taskQ = taskPipe #[0]
        self.resultQ = resultPipe# [1]
        self.scan_interval = scan_interval
        self.interval = interval
        self.bus = bus
        self.scanned = set()
        self.mysensors = dict()
        self.cycle_cnt = 0



    # ####################################w.init(b#
    # this part is running used in subprocess
    def do_scan(self, invoked_async = False):
        """ Initiate 1wire bus scanning,
            check existence in self.scanned
            join sens with mysensor
        """
        tmp_buf = set()
        self.cycle_cnt += 1
        for sens in self.ow.find(): 
            address = str(sens.address)[:14]
            tmp_buf.add(address)
            if not (address in self.scanned): # The sensor is scanned for a first time
                mysensor = self.mysensors.get(str(address))
                if not mysensor:
                    #if not found, create new one
                    mysensor = MySensorFabric(address, sens.type, self.interval)
                    if not mysensor: 
                        continue
                    self.mysensors[str(address)]=mysensor
                    # notify master process about new sensor - sensor is appended to mysensors HERE
                    self.resultQ.send((mysensor.address, mysensor.type, False))
                else:
                    logger.info("Sensor found ", address)
                mysensor.sens = sens
                self.scanned.add(address)

            elif invoked_async is True: # Reset planned scan time if invoked async
                mysensor = self.mysensors.get(str(address))
                if (mysensor) and (mysensor.lost is True):
                    mysensor.time = 0

        tmp_all = tmp_buf | self.scanned # Create union with all the sensors - both active and inactive
        inactives = (tmp_all - tmp_buf) # Get incative ones as difference
        for address in inactives: # All inactive sensors set as "lost"
            mysensor = self.mysensors.get(str(address))
            if mysensor and not mysensor.lost:
                mysensor.set_lost()
                self.resultQ.send((mysensor.circuit, mysensor.type, mysensor.lost))


    # This routine is invoked from the subprocess
    def do_command(self, cmd):
        command, address, value = cmd
        if command == OWCMD_INTERVAL:
            if address == 0: # Global interval change - for all connected sensors
                t1 = time.time()
                for mysensor in self.mysensors.values(): # Global change - for all sensors
                    mysensor.interval = value
                    mysensor.time = t1 + mysensor.calc_interval()
                    #self.resultQ.send((mysensor.address, mysensor.value, value))
            else: # Just for single sensor
                mysensor = self.mysensors.get(str(address))
                if mysensor:
                    mysensor.interval = value
                    mysensor.time = time.time() + mysensor.calc_interval()
                    #self.resultQ.send((mysensor.address, mysensor.value, value))

        elif command == OWCMD_SCAN:
            self.do_scan(invoked_async=True)
        elif command == OWCMD_SCAN_INTERVAL:
            self.scan_interval = value
        elif command == OWCMD_DEFAULT_INTERVAL:
            self.interval = value
        elif command == OWCMD_SET_PIO:
            mysensor = self.mysensors.get(str(address))
            if mysensor:
                pin, value = value
                mysensor.set_pio(pin, int(value))
        elif command == OWCMD_STOP:
            return 'DO_STOP'

    def run(self):
        """ Main loop of the subprocess
            Every scan_interval initiate bus scanning
            Peridocally scan 1wire sensors, else sleep
        """
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        self.ow = ow.Onewire(self.bus)

        logger.info("Entering OWW loop with PID {}".format(os.getpid()))

        while True: # If no sensors are in cache
            if self.scan_interval != 0: self.do_scan() # Do initial scan
            if len(self.mysensors) > 0:
                break

            if self.taskQ.poll(20): # Wait 20 second for CMD from main loop
                #CMD from main loop
                cmd = self.taskQ.recv()
                self.do_command(cmd)

        mysensor = min(self.mysensors.values(), key=lambda x: x.time) # Find sensor with min time (all se to 0 as default)

        scan_time = time.time() + self.scan_interval # Plan next scan
        do_stop = False

        while not do_stop: # "Main loop"
            if self.scan_interval != 0:
                t1 = time.time()
                if t1 >= scan_time: # Is time to scan the bus
                    self.do_scan()
                    scan_time = t1 + self.scan_interval # Plan next scan

            try:    # Read values from selected sensor
                t1 = time.time()
                mysensor.read_val_from_sens(mysensor.sens) # Read temperature from DS thermometer
                mysensor.lost = False # Readout was successful
                mysensor.readtime = t1 # Store last read time - UNUSED NOW, redundant to mysensor.time
                if self.resultQ:
                    # send measurement into result queue
                    self.resultQ.send((mysensor.address, mysensor.type, mysensor.value))
            except (ow.OnewireException, AttributeError):
                if not mysensor.lost: # Catch the edge
                    mysensor.set_lost()
                    self.resultQ.send((mysensor.address, mysensor.type, mysensor.lost)) # Send info about lost to the queue
            mysensor.time = t1 + mysensor.calc_interval()
            mysensor = min(self.mysensors.values(), key=lambda x: x.time)
            t1 = time.time()
            sleep_time = 0 if mysensor.time < t1 else mysensor.time - t1
            while (self.taskQ.poll(sleep_time) is True) and not do_stop:
                #commands from master
                try:
                    cmd = self.taskQ.recv()
                    do_stop = self.do_command(cmd) == 'DO_STOP'
                except Exception as E:
                    logger.critical("Error on do_command: %s", str(E))

                mysensor = min(self.mysensors.values(), key=lambda x: x.time)
                t1 = time.time()
                sleep_time = 0 if mysensor.time < t1 else mysensor.time - t1

        self.ow.finish()


class TQ:
    def poll(self, interval):
        if select.select([sys.stdin,],[],[],interval)[0]:
            return True
        return False

    def recv(self):
        #pass
        try:
            ret = json.loads(sys.stdin.readline())
            return ret
        except json.decoder.JSONDecodeError:
            raise Exception('Bad command on input')

class RQ:
    def send(self, *arg):
        sys.stdout.write(json.dumps(*arg)+'\n')
        sys.stdout.flush()



if __name__ == '__main__':
    import argparse

    a = argparse.ArgumentParser(prog='owprocess.py')
    #a.add_argument('progname')
    a.add_argument('-s','--scan_interval', help='interval of reading', default=300)
    a.add_argument('-i','--interval', help='interval of reading', default=3)
    a.add_argument('-2','--i2c', help='use this config file (default: homevok.conf from /etc or program directory)', 
                    default='/dev/i2c-1:ALL')
    args = a.parse_args()
    o = OwBusDriver(TQ(), RQ(), int(args.interval), int(args.scan_interval), bus='--i2c='+args.i2c)
    try:
        o.run()
    except (KeyboardInterrupt, BrokenPipeError):
        pass
